"""
    flask_zs.exceptions
    ~~~~~~~~~~~~

    flask-zs exceptions module

    :copyright: (c) 2018 by codeif.
    :license: MIT, see LICENSE for more details.
"""


class BaseCustomException(Exception):
    errcode = 1000
    errmsg = "Server Unkown Error."

    def __init__(self, errmsg=None, errcode=None, **kw):
        if errmsg:
            self.errmsg = errmsg
        if errcode is not None:
            self.errcode = errcode
        self.kw = kw

    def __str__(self):
        return "%d: %s" % (self.errcode, self.errmsg)

    def __repr__(self):
        return "<%s '%s'>" % (self.__class__.__name__, self)


class CustomException(BaseCustomException):
    pass


class NoError(CustomException):
    errcode = 0
    errmsg = "OK"



from flask import jsonify

from .exceptions import CustomException


def register_error_handlers(app):
    @app.errorhandler(CustomException)
    def custom_exception_handler(e):
        return jsonify(errcode=e.errcode, errmsg=e.errmsg, **e.kw)

    @app.errorhandler(401)
    @app.errorhandler(404)
    @app.errorhandler(405)
    def http_exception_handler(e):
        return jsonify(errcode=e.code, errmsg=e.name), e.code



from datetime import datetime
from sqlalchemy import Boolean, Column, DateTime, String
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.ext.hybrid import hybrid_property
from werkzeug.security import check_password_hash, generate_password_hash


class TimestampMixin:
    @declared_attr
    def created_at(cls):
        return Column(DateTime, default=datetime.now)

    @declared_attr
    def updated_at(cls):
        return Column(DateTime, default=datetime.now, onupdate=datetime.now)


class LoginMixin:
    @declared_attr
    def _password(cls):
        return Column("password", String(191), comment="login password")

    @declared_attr
    def login_allowed(cls):
        return Column(Boolean, server_default="0")

    @hybrid_property
    def password(self):
        return self._password
        # raise AttributeError('password is not a readable attribute')

    @password.setter
    def password(self, value):
        self._password = generate_password_hash(value)

    def check_password(self, password):
        if not self._password:
            return False
        return check_password_hash(self._password, password)



from .errorhandler import register_error_handlers
from .helpers import (
    BaseModel,
    CustomFlask,
    PaginationMixin,
    register_api,
    register_blueprints,
)
from .httpclient import HttpClient
from .response import make_errmsg

__all__ = (
    "BaseModel",
    "CustomFlask",
    "PaginationMixin",
    "register_api",
    "register_blueprints",
    "make_errmsg",
    "register_error_handlers",
    "HttpClient",
)



"""
    flask_zs.helpers
    ~~~~~~~~~~~~

    flask-zs helpers module

    :copyright: (c) 2018 by codeif.
    :license: MIT, see LICENSE for more details.
"""

import decimal
import importlib
import json
import uuid
from datetime import date, datetime, time

import requests
from flask import Blueprint, Flask
from werkzeug.utils import find_modules


def register_blueprints(
    app, import_path, bp_name="bp", include_packages=False, recursive=False
):
    """Register all Blueprint instances on the specified Flask application found
    in all modules for the specified package.

    :param app: the Flask application
    :param import_path: the dotted path for the package to find child modules.
    :param bp_name: Blueprint name in views.
    """
    for name in find_modules(
        import_path, include_packages=include_packages, recursive=recursive
    ):
        mod = importlib.import_module(name)
        bp = getattr(mod, bp_name, None)
        if isinstance(bp, Blueprint):
            app.register_blueprint(bp)


def register_api(bp, view_cls, endpoint, url, pk="item_id", pk_type="int"):
    """register restful api router

    :param bp: flask.BluePrint object
    :param view_cls: flask.views.View class
    :param endpoint: endpint
    :param url: url path, eg: /users/
    :param pk: entity id variable name
    :param pk_type:
        https://flask.palletsprojects.com/en/2.0.x/quickstart/#variable-rules
    """
    view_func = view_cls.as_view(endpoint)
    bp.add_url_rule(url, defaults={pk: None}, view_func=view_func, methods=["GET"])
    bp.add_url_rule(url, view_func=view_func, methods=["POST"])
    bp.add_url_rule(
        "{0}<{1}:{2}>".format(url, pk_type, pk),
        view_func=view_func,
        methods=["GET", "PUT", "DELETE", "PATCH"],
    )


class TodictMixin:
    _todict_include = None
    _todict_exclude = None
    _todict_simple = None

    def get_field_names(self):
        # for p in self.__mapper__.iterate_properties:
        #     yield p.key
        # _keys = self.__mapper__.c.keys()
        return [x.name for x in self.__table__.columns]

    def _get_todict_keys(self, include=None, exclude=None, only=None):
        if only:
            return only

        exclude_set = {"password", "created_at"}
        if self._todict_exclude:
            exclude_set.update(self._todict_exclude)
        if exclude:
            exclude_set.update(exclude)

        include_set = set()
        if self._todict_include:
            include_set.update(self._todict_include)
        if include:
            include_set.update(include)

        keys_set = set(self.get_field_names())
        keys_set.difference_update(exclude_set)
        keys_set.update(include_set)

        return keys_set

    def todict(self, include=None, exclude=None, only=None):
        keys = self._get_todict_keys(include, exclude, only)
        data = {key: getattr(self, key) for key in keys}
        return data or None

    def todict_simple(self):
        only = self._todict_simple or [
            x for x in self._get_todict_keys() if x in ["id", "name"]
        ]
        return self.todict(only=only)


class BaseModel(TodictMixin):
    pass


class JSONEncoder(json.JSONEncoder):
    def default(self, o):
        # See "Date Time String Format" in the ECMA-262 specification.
        if isinstance(o, set):
            return list(o)
        if isinstance(o, datetime):
            return o.isoformat(sep=" ", timespec="seconds")
        if isinstance(o, date):
            return o.isoformat()
        if isinstance(o, time):
            return o.isoformat(timespec="minutes")
        if isinstance(o, decimal.Decimal):
            return float(o)
        if isinstance(o, uuid.UUID):
            return str(o)
        if isinstance(o, TodictMixin):
            return o.todict_simple()
        else:
            return super().default(o)


def _iter_content(r, chunk_size=2 ** 12):
    while True:
        chunk = r.raw.read(chunk_size)
        if not chunk:
            break
        yield chunk


class CustomFlask(Flask):
    """使用自定义的JSONEncoder，并能处理view直接返回dict"""

    json_encoder = JSONEncoder

    def make_response(self, rv):
        if isinstance(rv, requests.Response):
            headers = rv.headers
            for key in [
                "Server",
                "Connection",
            ]:
                headers.pop(key, None)
            return self.response_class(
                _iter_content(rv), status=rv.status_code, headers=headers.items(),
            )

        return super().make_response(rv)


# Views Mixin
class PaginationMixin:
    def item_todict(self, item):
        return item.todict()

    def make_resp(self, query):
        return dict(items=[self.item_todict(x) for x in query])

    def make_pagination_resp(self, query):
        p = query.paginate(error_out=False)
        return dict(
            items=[self.item_todict(x) for x in p.items],
            pagination=dict(
                total=p.total, page=p.page, per_page=p.per_page, pages=p.pages
            ),
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['flask_zs', 'flask_zs.bin']

package_data = \
{'': ['*']}

install_requires = \
['Flask', 'SQLAlchemy', 'requests']

entry_points = \
{'console_scripts': ['collect-models = flask_zs.bin.collect_models:main']}

setup_kwargs = {
    'name': 'flask-zs',
    'version': '1.0.4',
    'description': 'A helpers for Flask.',
    'long_description': 'Helpers for Flask.\n====================\n\nHelpers for Flask. 使用示例 `codeif/flask-zs-template  <https://github.com/codeif/flask-zs-template>`_\n\n包含:\n\n- flask\n- sqlalchemy\n- requests\n\n\n安装\n----\n\n.. code-block:: sh\n\n    pip install flask-zs\n\n集中models\n-------------\n\n把models集中到models/__init__.py文件中(zsdemo为package name)::\n\n    PYTHONPATH=. collect-models [zsdemo]\n',
    'author': 'codeif',
    'author_email': 'me@codeif.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/codeif/flask-zs',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from flask import has_request_context, request
from requests import Session


class HttpClient:
    def __init__(
        self,
        app=None,
        base_url=None,
        headers=None,
        auth=None,
        user_agent=None,
        config_prefix="HTTP_CLIENT",
    ):
        self.base_url = base_url
        self.headers = headers
        self.auth = auth
        self.config_prefix = config_prefix
        self.user_agent = user_agent

        if app is not None:
            self.init_app(app)

    def init_app(self, app):
        if self.base_url is None:
            self.base_url = app.config[f"{self.config_prefix}_BASE_URL"]
        if self.headers is None:
            self.headers = app.config.get(f"{self.config_prefix}_HEADERS")
        if self.auth is None:
            self.auth = app.config.get(f"{self.config_prefix}_AUTH")
        if self.user_agent is None:
            self.auth = app.config.get(f"{self.config_prefix}_USER_AGENT")

    def create_session(self, headers=None, user_agent=None):
        """生成一个session, headers存在时覆盖默认haeders"""
        return ZSSession(
            base_url=self.base_url,
            headers=headers or self.headers,
            auth=self.auth,
            user_agent=user_agent or self.user_agent,
        )

    def request(self, method, path, **kwargs):
        with self.create_session() as session:
            return session.request(method, path, **kwargs)

    def get(self, path, **kwargs):
        return self.request("GET", path, **kwargs)

    def options(self, path, **kwargs):
        return self.request("OPTIONS", path, **kwargs)

    def head(self, path, **kwargs):
        return self.request("HEAD", path, **kwargs)

    def post(self, path, **kwargs):
        return self.request("POST", path, **kwargs)

    def put(self, path, **kwargs):
        return self.request("PUT", path, **kwargs)

    def patch(self, path, **kwargs):
        return self.request("PATCH", path, **kwargs)

    def delete(self, path, **kwargs):
        return self.request("DELETE", path, **kwargs)


class ZSSession(Session):
    def __init__(self, base_url, headers=None, auth=None, user_agent=None) -> None:
        super().__init__()
        self.stream = True

        self.base_url = base_url
        if headers:
            self.headers.update(headers)
        self.auth = auth
        if user_agent:
            self.headers["User-Agent"] = user_agent

    def request(self, method, path, **kwargs):
        url = self.base_url + path

        if has_request_context():
            accept_encoding = request.headers.get("Accept-Encoding")
            accept = request.headers.get("Accept")
            if accept_encoding:
                self.headers["Accept-Encoding"] = accept_encoding
            if accept:
                self.headers["Accept"] = accept

        return super().request(method, url, **kwargs)



import argparse
import inspect
import sys

from werkzeug.utils import find_modules, import_string

from .. import BaseModel


def find_classes(modname, import_path):
    cls_names = []
    mod = import_string(modname)
    for cls_name, cls_ in inspect.getmembers(mod, inspect.isclass):
        if issubclass(cls_, BaseModel):
            cls_names.append(cls_name)
    import_line = (
        f'from {modname.replace(import_path, "")} import {", ".join(cls_names)}'
    )
    return import_line, cls_names


def stdout_result(import_path):
    all_imports = []
    all_names = []
    for modname in find_modules(import_path):
        import_line, cls_names = find_classes(modname, import_path)
        if not cls_names:
            continue
        all_imports.append(import_line)
        all_names.extend(cls_names)

    if not all_imports:
        return

    import_str = "\n".join(all_imports)
    all_str = "\n    ".join([f'"{x}",' for x in sorted(all_names)])

    sys.stdout.write(
        f"""{import_str}

__all__ = [
    {all_str}
]
"""
    )


def main():
    parser = argparse.ArgumentParser(description="Collect all models to single file.")
    parser.add_argument(
        "project_name", help="current project name(package name), eg. zsdemo"
    )
    parsed_args = parser.parse_args()
    import_path = f"{parsed_args.project_name}.models"
    stdout_result(import_path)
